"use strict";
/*!
 * Copyright (c) Microsoft Corporation and contributors. All rights reserved.
 * Licensed under the MIT License.
 */
Object.defineProperty(exports, "__esModule", { value: true });
exports.SharedObjectHandle = exports.isISharedObjectHandle = void 0;
const internal_1 = require("@fluidframework/datastore/internal");
const runtime_utils_1 = require("@fluidframework/runtime-utils");
/**
 * Type guard for {@link ISharedObjectHandle}.
 * @internal
 */
function isISharedObjectHandle(handle) {
    return (0, runtime_utils_1.isFluidHandle)(handle) && typeof handle.bind === "function";
}
exports.isISharedObjectHandle = isISharedObjectHandle;
/**
 * Handle for a shared object.
 *
 * @remarks
 *
 * This object is used for already loaded (in-memory) shared objects and is used only for serialization purposes.
 *
 * De-serialization process goes through {@link @fluidframework/datastore#FluidObjectHandle}, and request flow:
 * {@link @fluidframework/datastore#FluidDataStoreRuntime.request} recognizes requests in the form of
 * '/\<shared object id\>' and loads shared object.
 */
class SharedObjectHandle extends internal_1.FluidObjectHandle {
    /**
     * Whether services have been attached for the associated shared object.
     */
    get isAttached() {
        return this.value.isAttached();
    }
    /**
     * Creates a new SharedObjectHandle.
     * @param value - The shared object this handle is for.
     * @param path - The id of the shared object. It is also the path to this object relative to the routeContext.
     * @param routeContext - The parent {@link @fluidframework/core-interfaces#IFluidHandleContext} that has a route
     * to this handle.
     */
    constructor(value, path, 
    // eslint-disable-next-line import/no-deprecated
    runtime) {
        super(value, path, runtime.IFluidHandleContext);
        this.value = value;
        this.runtime = runtime;
    }
    /**
     * Attaches all bound handles first (which may in turn attach further handles), then attaches this handle.
     * When attaching the handle, it registers the associated shared object.
     */
    attachGraph() {
        this.value.bindToContext();
        super.attachGraph();
    }
    bind(handle) {
        // We don't bind handles in staging mode to defer the attachment of any new objects
        // until we've exited staging mode. This way if we discard changes or a new handle is not present in the final
        // committed state, we will never end up attaching the discarded object.
        if (this.runtime.inStagingMode !== true) {
            super.bind(handle);
        }
    }
}
exports.SharedObjectHandle = SharedObjectHandle;
//# sourceMappingURL=handle.js.map